from sleplet.utils.vars import (
    PHI_MAX_DEFAULT,
    PHI_MIN_DEFAULT,
    THETA_MAX_DEFAULT,
    THETA_MIN_DEFAULT,
)


def is_polar_cap(
    phi_min: float, phi_max: float, theta_min: float, theta_max: float
) -> bool:
    """
    circular mask at the north pole
    """
    return (
        phi_min == PHI_MIN_DEFAULT
        and phi_max == PHI_MAX_DEFAULT
        and theta_min == THETA_MIN_DEFAULT
        and theta_max != THETA_MAX_DEFAULT
    )


def is_limited_lat_lon(
    phi_min: float, phi_max: float, theta_min: float, theta_max: float
) -> bool:
    """
    a region defined by angles, just need one to not be the default
    """
    return (
        not is_polar_cap(phi_min, phi_max, theta_min, theta_max)
        and phi_min != PHI_MIN_DEFAULT
        or phi_max != PHI_MAX_DEFAULT
        or theta_min != THETA_MIN_DEFAULT
    )


def is_ergodic(j_min: int, *, j: int = 0) -> bool:
    """
    computes whether the function follows ergodicity

    ergodicity fails for J_min = 0, because the scaling function will only
    cover f00. Hence <flm flm*> will be 0 in that case and the scaling
    coefficients will all be the same. So, if we do have J_min=0, we take the
    variance over all realisations instead (of course, we then won't have a
    standard deviation to compare it to the theoretical variance).
    """
    return j_min != 0 or j != 0



import numpy as np

from sleplet.utils.bool_methods import is_ergodic, is_limited_lat_lon, is_polar_cap
from sleplet.utils.vars import PHI_MAX_DEFAULT, PHI_MIN_DEFAULT, THETA_MIN_DEFAULT

J = 2
J_MIN = 0
PHI_0 = np.pi / 6
PHI_1 = np.pi / 3
THETA_0 = np.pi / 6
THETA_1 = np.pi / 3
THETA_MAX = 2 * np.pi / 9


def test_bool_polar_cap() -> None:
    """
    verifies that one is case is a polar cap and one isn't
    """
    assert is_polar_cap(PHI_MIN_DEFAULT, PHI_MAX_DEFAULT, THETA_MIN_DEFAULT, THETA_MAX)
    assert not is_polar_cap(PHI_0, PHI_1, THETA_0, THETA_1)


def test_bool_lim_lat_lon() -> None:
    """
    verifies that one is case is a polar cap and one isn't
    """
    assert is_limited_lat_lon(PHI_0, PHI_1, THETA_0, THETA_1)
    assert not is_limited_lat_lon(
        PHI_MIN_DEFAULT, PHI_MAX_DEFAULT, THETA_MIN_DEFAULT, THETA_MAX
    )


def test_bool_erodicity() -> None:
    """
    verifies that a function follows ergodicity
    """
    assert not is_ergodic(J_MIN)
    assert is_ergodic(J_MIN, j=J)



from pathlib import Path

import numpy as np
import seaborn as sns
from matplotlib import pyplot as plt
from numpy import typing as npt

from sleplet.functions.flm.earth import Earth
from sleplet.slepian.slepian_region.slepian_polar_cap import SlepianPolarCap
from sleplet.utils.plot_methods import save_plot
from sleplet.utils.region import Region
from sleplet.utils.slepian_methods import choose_slepian_method, slepian_forward

L = 16
THETA_MAX = 40

fig_path = Path(__file__).resolve().parents[2] / "src" / "sleplet" / "figures"
sns.set(context="paper")


def _earth_region_harmonic_coefficients(
    L: int, theta_max: int
) -> npt.NDArray[np.float_]:
    """
    harmonic coefficients of the Earth for the polar cap region
    """
    region = Region(theta_max=np.deg2rad(theta_max))
    earth = Earth(L, region=region)
    coefficients = np.abs(earth.coefficients)
    coefficients[::-1].sort()
    return coefficients


def _earth_region_slepian_coefficients(
    L: int, theta_max: int
) -> npt.NDArray[np.float_]:
    """
    computes the Slepian coefficients
    """
    region = Region(theta_max=np.deg2rad(theta_max))
    earth = Earth(L, region=region)
    slepian = choose_slepian_method(L, region)
    return np.abs(slepian_forward(L, slepian, flm=earth.coefficients))


def main() -> None:
    """
    creates a plot of Slepian coefficients against rank
    """
    N = SlepianPolarCap(L, np.deg2rad(THETA_MAX)).N
    flm = _earth_region_harmonic_coefficients(L, THETA_MAX)[:N]
    f_p = np.sort(_earth_region_slepian_coefficients(L, THETA_MAX))[::-1]
    ax = plt.gca()
    sns.scatterplot(x=range(N), y=f_p, ax=ax, label="slepian", linewidth=0, marker="*")
    sns.scatterplot(x=range(N), y=flm, ax=ax, label="harmonic", linewidth=0, marker=".")
    ax.set_xlabel("coefficients")
    ax.set_ylabel("magnitude")
    save_plot(fig_path, f"fp_earth_polar{THETA_MAX}_L{L}")


if __name__ == "__main__":
    main()



import numpy as np
import pyssht as ssht
from numpy.random import default_rng

from sleplet import logger
from sleplet.functions.flm.axisymmetric_wavelets import AxisymmetricWavelets
from sleplet.utils.bool_methods import is_ergodic
from sleplet.utils.harmonic_methods import compute_random_signal
from sleplet.utils.vars import RANDOM_SEED, SAMPLING_SCHEME
from sleplet.utils.wavelet_methods import (
    axisymmetric_wavelet_forward,
    compute_wavelet_covariance,
)

B = 3
J_MIN = 2
L = 128


def axisymmetric_wavelet_covariance(
    L: int, B: int, j_min: int, *, runs: int = 100, var_flm: float = 1
) -> None:
    """
    compute theoretical covariance of wavelet coefficients

    the covariance <Wj(omega)Wj*(omega)> is given by the following expression:
    sigma^2 Sum(l,0) |Psi^j_l0|^2

    where sigma^2 is the variance of the harmonic coefficients and Psi^j_l0
    are the harmonic coefficients of the j-th wavelet

    a similar expression applies for the scaling function coefficients

    should we use the actual variance of each realisation instead?
    """
    logger.info(f"L={L}, B={B}, j_min={j_min}")

    # compute wavelets
    aw = AxisymmetricWavelets(L, B=B, j_min=j_min)

    # theoretical covariance
    covar_theory = compute_wavelet_covariance(L, aw.wavelets, var_signal=var_flm)

    # initialise matrix
    covar_runs_shape = (runs, *covar_theory.shape)
    covar_data = np.zeros(covar_runs_shape, dtype=np.complex_)

    # set seed
    rng = default_rng(RANDOM_SEED)

    for i in range(runs):
        logger.info(f"start run: {i+1}/{runs}")

        # Generate normally distributed random complex signal
        flm = compute_random_signal(L, rng, var_signal=var_flm)

        # compute wavelet coefficients
        wlm = axisymmetric_wavelet_forward(L, flm, aw.wavelets)

        # compute covariance from data
        for j, coefficient in enumerate(wlm):
            f_wav_j = ssht.inverse(coefficient, L, Method=SAMPLING_SCHEME)
            covar_data[i, j] = (
                f_wav_j.var() if is_ergodic(j_min, j=j) else f_wav_j[0, 0]
            )

    # compute mean and variance
    mean_covar_data = covar_data.mean(axis=0)
    std_covar_data = covar_data.std(axis=0)

    # override for scaling function
    if not is_ergodic(j_min):
        mean_covar_data[0] = covar_data[0].var()

    # ensure reality
    mean_covar_data = np.abs(mean_covar_data)

    # compute errors
    error_absolute = np.abs(mean_covar_data - covar_theory)
    error_in_std = error_absolute / std_covar_data

    # report errors
    for j in range(len(aw.wavelets)):
        message = (
            f"error in std: {error_in_std[j]:e}"
            if is_ergodic(j_min, j=j)
            else f"absolute error: {error_absolute[j]:e}"
        )
        logger.info(f"axisymmetric wavelet covariance {j}: '{message}'")


if __name__ == "__main__":
    axisymmetric_wavelet_covariance(L, B, J_MIN)



from plotly.graph_objs import Layout
from plotly.graph_objs.layout import Margin, Scene
from plotly.graph_objs.layout.scene import Camera, XAxis, YAxis, ZAxis
from plotly.graph_objs.layout.scene.camera import Center, Eye

_axis = {
    "title": "",
    "showgrid": False,
    "zeroline": False,
    "ticks": "",
    "showticklabels": False,
    "showbackground": False,
}


def create_camera(
    x_eye: float,
    y_eye: float,
    z_eye: float,
    zoom: float,
    *,
    x_center: float = 0,
    y_center: float = 0,
    z_center: float = 0,
) -> Camera:
    """
    creates default camera view with a zoom factor
    """
    return Camera(
        eye=Eye(x=x_eye / zoom, y=y_eye / zoom, z=z_eye / zoom),
        center=Center(x=x_center, y=y_center, z=z_center),
    )


def create_layout(camera: Camera, *, annotations: list[dict] | None = None) -> Layout:
    """
    default plotly layout
    """
    return Layout(
        scene=Scene(
            dragmode="orbit",
            camera=camera,
            xaxis=XAxis(_axis),
            yaxis=YAxis(_axis),
            zaxis=ZAxis(_axis),
            annotations=annotations,
        ),
        margin=Margin(l=0, r=0, b=0, t=0),
        paper_bgcolor="rgba(0,0,0,0)",
        plot_bgcolor="rgba(0,0,0,0)",
    )


def create_tick_mark(
    fmin: float, fmax: float, *, amplitude: float | None = None
) -> float:
    """
    creates tick mark to use when using a non-normalised plot
    """
    return amplitude if amplitude is not None else max(abs(fmin), abs(fmax))


def create_colour_bar(
    tick_mark: float,
    *,
    normalise: bool,
    bar_len: float = 0.94,
    bar_pos: float = 0.97,
    font_size: int = 38,
) -> dict:
    """
    default plotly colour bar
    """
    return {
        "x": bar_pos,
        "len": bar_len,
        "nticks": 2 if normalise else None,
        "tickfont": {"color": "#666666", "size": font_size},
        "tickformat": None if normalise else "+.1e",
        "tick0": None if normalise else -tick_mark,
        "dtick": None if normalise else tick_mark,
    }



import numpy as np
import pyssht as ssht
from numpy import typing as npt

from sleplet.functions.coefficients import Coefficients
from sleplet.functions.f_p import F_P
from sleplet.functions.flm.axisymmetric_wavelets import AxisymmetricWavelets
from sleplet.functions.fp.slepian_wavelets import SlepianWavelets
from sleplet.meshes.mesh_coefficients import MeshCoefficients
from sleplet.meshes.slepian_coefficients.mesh_slepian_wavelets import (
    MeshSlepianWavelets,
)
from sleplet.utils.harmonic_methods import rotate_earth_to_south_america
from sleplet.utils.noise import (
    compute_sigma_j,
    compute_sigma_noise,
    compute_slepian_mesh_sigma_j,
    compute_slepian_sigma_j,
    compute_snr,
    harmonic_hard_thresholding,
    slepian_function_hard_thresholding,
    slepian_mesh_hard_thresholding,
    slepian_wavelet_hard_thresholding,
)
from sleplet.utils.slepian_methods import slepian_inverse, slepian_mesh_inverse
from sleplet.utils.vars import SAMPLING_SCHEME
from sleplet.utils.wavelet_methods import (
    axisymmetric_wavelet_forward,
    axisymmetric_wavelet_inverse,
    slepian_wavelet_forward,
    slepian_wavelet_inverse,
)


def denoising_axisym(
    signal: Coefficients,
    noised_signal: Coefficients,
    axisymmetric_wavelets: AxisymmetricWavelets,
    snr_in: float,
    n_sigma: int,
    *,
    rotate_to_south_america: bool = False,
) -> tuple[npt.NDArray[np.complex_], float | None, float]:
    """
    reproduce the denoising demo from s2let paper
    """
    # compute wavelet coefficients
    w = axisymmetric_wavelet_forward(
        signal.L, noised_signal.coefficients, axisymmetric_wavelets.wavelets
    )

    # compute wavelet noise
    sigma_j = compute_sigma_j(
        signal.coefficients, axisymmetric_wavelets.wavelets[1:], snr_in
    )

    # hard thresholding
    w_denoised = harmonic_hard_thresholding(signal.L, w, sigma_j, n_sigma)

    # wavelet synthesis
    flm = axisymmetric_wavelet_inverse(
        signal.L, w_denoised, axisymmetric_wavelets.wavelets
    )

    # compute SNR
    denoised_snr = compute_snr(
        signal.coefficients, flm - signal.coefficients, "Harmonic"
    )

    # rotate to South America
    flm = (
        rotate_earth_to_south_america(flm, signal.L) if rotate_to_south_america else flm
    )

    f = ssht.inverse(flm, signal.L, Method=SAMPLING_SCHEME)
    return f, noised_signal.snr, denoised_snr


def denoising_slepian_wavelet(
    signal: Coefficients,
    noised_signal: Coefficients,
    slepian_wavelets: SlepianWavelets,
    snr_in: float,
    n_sigma: int,
) -> npt.NDArray[np.complex_]:
    """
    denoising demo using Slepian wavelets
    """
    # compute wavelet coefficients
    w = slepian_wavelet_forward(
        noised_signal.coefficients,
        slepian_wavelets.wavelets,
        slepian_wavelets.slepian.N,
    )

    # compute wavelet noise
    sigma_j = compute_slepian_sigma_j(
        signal.L,
        signal.coefficients,
        slepian_wavelets.wavelets,
        snr_in,
        slepian_wavelets.slepian,
    )

    # hard thresholding
    w_denoised = slepian_wavelet_hard_thresholding(
        signal.L, w, sigma_j, n_sigma, slepian_wavelets.slepian
    )

    # wavelet synthesis
    f_p = slepian_wavelet_inverse(
        w_denoised, slepian_wavelets.wavelets, slepian_wavelets.slepian.N
    )

    # compute SNR
    compute_snr(signal.coefficients, f_p - signal.coefficients, "Slepian")

    return slepian_inverse(f_p, signal.L, slepian_wavelets.slepian)


def denoising_slepian_function(
    signal: F_P,
    noised_signal: F_P,
    snr_in: float,
    n_sigma: int,
) -> npt.NDArray[np.complex_]:
    """
    denoising demo using Slepian function
    """
    # compute Slepian noise
    sigma_noise = compute_sigma_noise(
        signal.coefficients, snr_in, denominator=signal.L**2
    )

    # hard thresholding
    f_p = slepian_function_hard_thresholding(
        signal.L, noised_signal.coefficients, sigma_noise, n_sigma, signal.slepian
    )

    # compute SNR
    compute_snr(signal.coefficients, f_p - signal.coefficients, "Slepian")

    return slepian_inverse(f_p, signal.L, signal.slepian)


def denoising_mesh_slepian(
    signal: MeshCoefficients,
    noised_signal: MeshCoefficients,
    mesh_slepian_wavelets: MeshSlepianWavelets,
    snr_in: float,
    n_sigma: int,
) -> npt.NDArray[np.complex_ | np.float_]:
    """
    denoising demo using Slepian wavelets
    """
    # compute wavelet coefficients
    w = slepian_wavelet_forward(
        noised_signal.coefficients,
        mesh_slepian_wavelets.wavelets,
        mesh_slepian_wavelets.mesh_slepian.N,
    )

    # compute wavelet noise
    sigma_j = compute_slepian_mesh_sigma_j(
        mesh_slepian_wavelets.mesh_slepian,
        signal.coefficients,
        mesh_slepian_wavelets.wavelets,
        snr_in,
    )

    # hard thresholding
    w_denoised = slepian_mesh_hard_thresholding(
        mesh_slepian_wavelets.mesh_slepian, w, sigma_j, n_sigma
    )

    # wavelet synthesis
    f_p = slepian_wavelet_inverse(
        w_denoised, mesh_slepian_wavelets.wavelets, mesh_slepian_wavelets.mesh_slepian.N
    )

    # compute SNR
    compute_snr(signal.coefficients, f_p - signal.coefficients, "Slepian")

    return slepian_mesh_inverse(mesh_slepian_wavelets.mesh_slepian, f_p)



import numpy as np
from numpy import typing as npt
from pydantic.dataclasses import dataclass

from sleplet.functions.f_p import F_P
from sleplet.functions.fp.slepian_africa import SlepianAfrica
from sleplet.utils.noise import compute_snr, create_slepian_noise
from sleplet.utils.region import Region
from sleplet.utils.string_methods import convert_camel_case_to_snake_case, filename_args
from sleplet.utils.validation import Validation


@dataclass(config=Validation, kw_only=True)
class SlepianNoiseAfrica(F_P):
    SNR: float = -10

    def __post_init_post_parse__(self) -> None:
        super().__post_init_post_parse__()
        if isinstance(self.region, Region) and self.region.name_ending != "africa":
            raise RuntimeError("Slepian region selected must be 'africa'")

    def _create_coefficients(self) -> npt.NDArray[np.complex_ | np.float_]:
        sa = SlepianAfrica(self.L, region=self.region, smoothing=self.smoothing)
        noise = create_slepian_noise(self.L, sa.coefficients, self.slepian, self.SNR)
        compute_snr(sa.coefficients, noise, "Slepian")
        return noise

    def _create_name(self) -> str:
        return (
            f"{convert_camel_case_to_snake_case(self.__class__.__name__)}"
            f"{filename_args(self.SNR, 'snr')}"
        )

    def _set_reality(self) -> bool:
        return False

    def _set_spin(self) -> int:
        return 0

    def _setup_args(self) -> None:
        if isinstance(self.extra_args, list):
            num_args = 1
            if len(self.extra_args) != num_args:
                raise ValueError(f"The number of extra arguments should be {num_args}")
            self.SNR = self.extra_args[0]



import numpy as np
from numpy.testing import assert_equal

from sleplet.utils.string_methods import (
    angle_as_degree,
    filename_angle,
    filename_args,
    multiples_of_pi,
    wavelet_ending,
)

J_MIN = 0
PHI_0 = np.pi / 6
PHI_1 = np.pi / 3
THETA_MAX = 2 * np.pi / 9


def test_add_extra_args_to_filename() -> None:
    """
    test extra args added to filename
    """
    arguments = [0, 123, 0.004]
    output = ["_0a", "_123a", "_1a250"]
    for c, arg in enumerate(arguments):
        assert_equal(filename_args(arg, "a"), output[c])


def test_add_angle_to_filename() -> None:
    """
    tests angle added to filename
    """
    arguments = [(0.75, 0.125, 0), (0.1, 0.2, 0.3), (0.9, 0, 0), (0, 0.8, 0), (0, 0, 0)]
    output = [
        "alpha3pi4_beta1pi8",
        "alpha1pi10_beta1pi5_gamma3pi10",
        "alpha9pi10_beta0",
        "alpha0_beta4pi5",
        "alpha0_beta0",
    ]
    for c, (alpha, beta, gamma) in enumerate(arguments):
        assert_equal(filename_angle(alpha, beta, gamma), output[c])


def test_print_multiple_of_pi() -> None:
    """
    tests that the pi prefix is added
    """
    arguments = [0, 1, 2, 2.5]
    output = ["0\u03C0", "\u03C0", "2\u03C0", "2\u03C0"]
    for c, arg in enumerate(arguments):
        assert_equal(multiples_of_pi(arg * np.pi), output[c])


def test_convert_angle_to_degrees() -> None:
    """
    verifies angles is converted to degree
    """
    arguments = [PHI_0, PHI_1, THETA_MAX]
    output = [30, 60, 40]
    for c, arg in enumerate(arguments):
        assert_equal(angle_as_degree(arg), output[c])


def test_add_to_wavelet_name() -> None:
    """
    test that the correct ending for wavelets is added
    """
    arguments = [None, 0, 1]
    output = ["_scaling", "_0j", "_1j"]
    for c, arg in enumerate(arguments):
        assert_equal(wavelet_ending(J_MIN, arg), output[c])



import numpy as np
from igl import per_vertex_normals
from numpy import typing as npt
from pydantic.dataclasses import dataclass

from sleplet.meshes.mesh_harmonic_coefficients import MeshHarmonicCoefficients
from sleplet.utils.harmonic_methods import mesh_forward
from sleplet.utils.validation import Validation


@dataclass(config=Validation)
class MeshField(MeshHarmonicCoefficients):
    def __post_init_post_parse__(self) -> None:
        super().__post_init_post_parse__()

    def _create_coefficients(self) -> npt.NDArray[np.complex_ | np.float_]:
        """
        compute field on the vertices of the mesh
        """
        field = per_vertex_normals(self.mesh.vertices, self.mesh.faces)[:, 1]
        return mesh_forward(self.mesh, field)

    def _create_name(self) -> str:
        return f"{self.mesh.name}_field"

    def _setup_args(self) -> None:
        if isinstance(self.extra_args, list):
            raise AttributeError(
                f"{self.__class__.__name__} does not support extra arguments"
            )



import numpy as np
from numpy import typing as npt
from pydantic.dataclasses import dataclass

from sleplet.functions.f_lm import F_LM
from sleplet.utils.string_methods import convert_camel_case_to_snake_case
from sleplet.utils.validation import Validation


@dataclass(config=Validation)
class Identity(F_LM):
    def __post_init_post_parse__(self) -> None:
        super().__post_init_post_parse__()

    def _create_coefficients(self) -> npt.NDArray[np.complex_ | np.float_]:
        return np.ones(self.L**2, dtype=np.complex_)

    def _create_name(self) -> str:
        return convert_camel_case_to_snake_case(self.__class__.__name__)

    def _set_reality(self) -> bool:
        return True

    def _set_spin(self) -> int:
        return 0

    def _setup_args(self) -> None:
        if isinstance(self.extra_args, list):
            raise AttributeError(
                f"{self.__class__.__name__} does not support extra arguments"
            )



import numpy as np
import pyssht as ssht
from numpy import typing as npt
from pydantic.dataclasses import dataclass

from sleplet import logger
from sleplet.functions.f_p import F_P
from sleplet.utils.slepian_methods import compute_s_p_omega_prime
from sleplet.utils.string_methods import convert_camel_case_to_snake_case
from sleplet.utils.validation import Validation
from sleplet.utils.vars import SAMPLING_SCHEME


@dataclass(config=Validation)
class SlepianDiracDelta(F_P):
    def __post_init_post_parse__(self) -> None:
        super().__post_init_post_parse__()

    def _create_coefficients(self) -> npt.NDArray[np.complex_ | np.float_]:
        self._compute_angles()
        return compute_s_p_omega_prime(
            self.L, self.alpha, self.beta, self.slepian
        ).conj()

    def _create_name(self) -> str:
        return (
            f"{convert_camel_case_to_snake_case(self.__class__.__name__)}"
            f"_{self.slepian.region.name_ending}"
        )

    def _set_reality(self) -> bool:
        return False

    def _set_spin(self) -> int:
        return 0

    def _setup_args(self) -> None:
        if isinstance(self.extra_args, list):
            raise AttributeError(
                f"{self.__class__.__name__} does not support extra arguments"
            )

    def _compute_angles(self) -> None:
        """
        computes alpha/beta if not provided
        """
        thetas, phis = ssht.sample_positions(self.L, Grid=True, Method=SAMPLING_SCHEME)
        sp = ssht.inverse(self.slepian.eigenvectors[0], self.L, Method=SAMPLING_SCHEME)
        idx = tuple(np.argwhere(sp == sp.max())[0])
        self.alpha = phis[idx]
        self.beta = thetas[idx]
        logger.info(
            f"angles: (alpha, beta) = ({self.alpha/np.pi:.5f},{self.beta/np.pi:.5f})"
        )
        logger.info(f"grid point: (alpha, beta) = ({self.alpha:e},{self.beta:e})")



from sleplet.functions.flm.axisymmetric_wavelets import AxisymmetricWavelets
from sleplet.functions.flm.earth import Earth
from sleplet.plotting.create_plot_sphere import Plot
from sleplet.utils.denoising import denoising_axisym
from sleplet.utils.plot_methods import find_max_amplitude

B = 2
J_MIN = 0
L = 128
N_SIGMA = 3
NORMALISE = False
SNR_IN = 10


def main() -> None:
    """
    reproduce the denoising demo from s2let paper
    """
    # create map & noised map
    fun = Earth(L)
    fun_noised = Earth(L, noise=SNR_IN)

    # create wavelets
    aw = AxisymmetricWavelets(L, B=B, j_min=J_MIN)

    # fix amplitude
    amplitude = find_max_amplitude(fun)

    f, _, _ = denoising_axisym(
        fun, fun_noised, aw, SNR_IN, N_SIGMA, rotate_to_south_america=True
    )
    name = f"{fun.name}_denoised_axisym"
    Plot(f, L, name, amplitude=amplitude, normalise=NORMALISE).execute()


if __name__ == "__main__":
    main()



import numpy as np
from numpy import typing as npt
from pydantic.dataclasses import dataclass

from sleplet.meshes.harmonic_coefficients.mesh_field import MeshField
from sleplet.meshes.mesh_slepian_coefficients import MeshSlepianCoefficients
from sleplet.utils.slepian_methods import slepian_mesh_forward
from sleplet.utils.validation import Validation


@dataclass(config=Validation)
class MeshSlepianField(MeshSlepianCoefficients):
    def __post_init_post_parse__(self) -> None:
        super().__post_init_post_parse__()

    def _create_coefficients(self) -> npt.NDArray[np.complex_ | np.float_]:
        """
        compute field on the vertices of the mesh
        """
        mf = MeshField(self.mesh, region=True)
        return slepian_mesh_forward(
            self.mesh_slepian,
            u_i=mf.coefficients,
        )

    def _create_name(self) -> str:
        return f"slepian_{self.mesh.name}_field"

    def _setup_args(self) -> None:
        if isinstance(self.extra_args, list):
            raise AttributeError(
                f"{self.__class__.__name__} does not support extra arguments"
            )



from dataclasses import KW_ONLY

import numpy as np
import pyssht as ssht
from numpy import typing as npt
from pydantic.dataclasses import dataclass

from sleplet import logger
from sleplet.slepian.slepian_functions import SlepianFunctions
from sleplet.utils.integration_methods import (
    calc_integration_weight,
    integrate_region_sphere,
    integrate_whole_sphere,
)
from sleplet.utils.validation import Validation
from sleplet.utils.vars import SAMPLING_SCHEME


@dataclass(config=Validation)
class SlepianDecomposition:
    L: int
    slepian: SlepianFunctions
    _: KW_ONLY
    f: npt.NDArray[np.complex_] | None = None
    flm: npt.NDArray[np.complex_ | np.float_] | None = None
    mask: npt.NDArray[np.float_] | None = None

    def __post_init_post_parse__(self) -> None:
        self._detect_method()

    def decompose(self, rank: int) -> complex:
        """
        decompose the signal into its Slepian coefficients via the given method
        """
        self._validate_rank(rank)

        match self.method:
            case "harmonic_sum":
                return self._harmonic_sum(rank)
            case "integrate_sphere":
                return self._integrate_sphere(rank)
            case "integrate_region":
                return self._integrate_region(rank)
            case _:
                raise ValueError(f"'{self.method}' is not a valid method")

    def decompose_all(self, n_coefficients: int) -> npt.NDArray[np.complex_]:
        """
        decompose all ranks of the Slepian coefficients
        """
        coefficients = np.zeros(n_coefficients, dtype=np.complex_)
        for rank in range(n_coefficients):
            coefficients[rank] = self.decompose(rank)
        return coefficients

    def _integrate_region(self, rank: int) -> complex:
        r"""
        f_{p} =
        \frac{1}{\lambda_{p}}
        \int\limits_{R} \dd{\Omega(\omega)}
        f(\omega) \overline{S_{p}(\omega)}
        """
        assert isinstance(self.mask, np.ndarray)  # noqa: S101
        assert isinstance(self.f, np.ndarray)  # noqa: S101
        s_p = ssht.inverse(
            self.slepian.eigenvectors[rank], self.L, Method=SAMPLING_SCHEME
        )
        weight = calc_integration_weight(self.L)
        integration = integrate_region_sphere(self.mask, weight, self.f, s_p.conj())
        return integration / self.slepian.eigenvalues[rank]

    def _integrate_sphere(self, rank: int) -> complex:
        r"""
        f_{p} =
        \int\limits_{S^{2}} \dd{\Omega(\omega)}
        f(\omega) \overline{S_{p}(\omega)}
        """
        assert isinstance(self.f, np.ndarray)  # noqa: S101
        s_p = ssht.inverse(
            self.slepian.eigenvectors[rank], self.L, Method=SAMPLING_SCHEME
        )
        weight = calc_integration_weight(self.L)
        return integrate_whole_sphere(weight, self.f, s_p.conj())

    def _harmonic_sum(self, rank: int) -> complex:
        r"""
        f_{p} =
        \sum\limits_{\ell=0}^{L^{2}}
        \sum\limits_{m=-\ell}^{\ell}
        f_{\ell m} (S_{p})_{\ell m}^{*}
        """
        return (self.flm * self.slepian.eigenvectors[rank].conj()).sum()

    def _detect_method(self) -> None:
        """
        detects what method is used to perform the decomposition
        """
        if isinstance(self.flm, np.ndarray):
            logger.info("harmonic sum method selected")
            self.method = "harmonic_sum"
        elif isinstance(self.f, np.ndarray) and not isinstance(self.mask, np.ndarray):
            logger.info("integrating the whole sphere method selected")
            self.method = "integrate_sphere"
        elif isinstance(self.f, np.ndarray):
            logger.info("integrating a region on the sphere method selected")
            self.method = "integrate_region"
        else:
            raise RuntimeError(
                "need to pass one off harmonic coefficients, real pixels "
                "or real pixels with a mask"
            )

    def _validate_rank(self, rank: int) -> None:
        """
        checks the requested rank is valid
        """
        if not isinstance(rank, int):
            raise TypeError("rank should be an integer")
        if rank < 0:
            raise ValueError("rank cannot be negative")
        if rank >= self.L**2:
            raise ValueError(f"rank should be less than {self.L**2}")



import os

import numpy as np

from sleplet import logger

AFRICA_ALPHA = np.deg2rad(44)
AFRICA_BETA = np.deg2rad(87)
AFRICA_GAMMA = np.deg2rad(341)
AFRICA_RANGE = np.deg2rad(41)
ALPHA_DEFAULT = 0.75
ANNOTATION_COLOUR = "gold"
ARROW_STYLE = {
    "arrowhead": 0,
    "arrowside": "start",
    "ax": 4,
    "ay": 4,
    "startarrowsize": 0.5,
    "startarrowhead": 6,
}
BETA_DEFAULT = 0.125
MESH_CBAR_FONT_SIZE = 32
MESH_CBAR_LEN = 0.95
MESH_UNSEEN = -1e5  # kaleido bug
NCPU = int(os.getenv("NCPU", "4"))
PHI_0 = np.pi
PHI_MAX_DEFAULT = 2 * np.pi
PHI_MAX = int(os.getenv("PHI_MAX", "360"))
PHI_MIN_DEFAULT = 0.0
PHI_MIN = int(os.getenv("PHI_MIN", "0"))
POLAR_GAP = os.getenv("POLAR_GAP", "False").lower() == "true"
RANDOM_SEED = 30
SAMPLES = 2
SAMPLING_SCHEME = "MWSS"
SLEPIAN_MASK = os.getenv("SLEPIAN_MASK", "south_america")
SMOOTHING = 2
SOUTH_AMERICA_ALPHA = np.deg2rad(54)
SOUTH_AMERICA_BETA = np.deg2rad(108)
SOUTH_AMERICA_GAMMA = np.deg2rad(63)
SOUTH_AMERICA_RANGE = np.deg2rad(40)
SPHERE_UNSEEN = -1.56e30
THETA_0 = 0.0
THETA_MAX_DEFAULT = np.pi
THETA_MAX = int(os.getenv("THETA_MAX", "180"))
THETA_MIN_DEFAULT = 0
THETA_MIN = int(os.getenv("THETA_MIN", "0"))
ZENODO_DATA_DOI = "10.5281/zenodo.7767698"

logger.info(
    "Environment variables set as "
    f"NCPU={NCPU}, "
    f"POLAR_GAP={POLAR_GAP}, "
    f"THETA_MAX={THETA_MAX}, "
    f"THETA_MIN={THETA_MIN}, "
    f"PHI_MAX={PHI_MAX}, "
    f"PHI_MIN={PHI_MIN}, "
    f"SLEPIAN_MASK={SLEPIAN_MASK}."
)



from pathlib import Path

import numpy as np
import pyssht as ssht
from numpy import typing as npt

from sleplet import logger
from sleplet.data.other.earth.create_earth_flm import create_flm
from sleplet.data.setup_pooch import find_on_pooch_then_local
from sleplet.meshes.classes.mesh import Mesh
from sleplet.utils.harmonic_methods import (
    mesh_forward,
    mesh_inverse,
    rotate_earth_to_africa,
    rotate_earth_to_south_america,
)
from sleplet.utils.region import Region
from sleplet.utils.vars import (
    AFRICA_RANGE,
    PHI_MAX,
    PHI_MIN,
    POLAR_GAP,
    SAMPLING_SCHEME,
    SLEPIAN_MASK,
    SOUTH_AMERICA_RANGE,
    THETA_MAX,
    THETA_MIN,
)

_data_path = Path(__file__).resolve().parents[1] / "data"


def create_mask_region(L: int, region: Region) -> npt.NDArray[np.float_]:
    """
    creates a mask of a region of interested, the output will be based
    on the value of the provided L. The mask could be either:
    * polar cap - if theta_max provided
    * limited latitude longitude - if one of theta_min, theta_max,
                                   phi_min or phi_max is provided
    * arbitrary - just checks the shape of the input mask
    """
    thetas, phis = ssht.sample_positions(L, Grid=True, Method=SAMPLING_SCHEME)

    match region.region_type:
        case "arbitrary":
            logger.info("loading and checking shape of provided mask")
            name = f"{region.mask_name}_L{L}.npy"
            mask = _load_mask(L, name)
            assert mask.shape == thetas.shape, (  # noqa: S101
                f"mask {name} has shape {mask.shape} which does not match "
                f"the provided L={L}, the shape should be {thetas.shape}"
            )

        case "lim_lat_lon":
            logger.info("creating limited latitude longitude mask")
            mask = (
                (thetas >= region.theta_min)
                & (thetas <= region.theta_max)
                & (phis >= region.phi_min)
                & (phis <= region.phi_max)
            )

        case "polar":
            logger.info("creating polar cap mask")
            mask = thetas <= region.theta_max
            if region.gap:
                logger.info("creating polar gap mask")
                mask += thetas >= np.pi - region.theta_max
    return mask


def _load_mask(L: int, mask_name: str) -> npt.NDArray[np.float_]:
    """
    attempts to read the mask from the config file
    """
    mask = find_on_pooch_then_local(f"slepian_masks_{mask_name}")
    return create_mask(L, mask_name) if mask is None else np.load(mask)


def ensure_masked_flm_bandlimited(
    flm: npt.NDArray[np.complex_], L: int, region: Region, *, reality: bool, spin: int
) -> npt.NDArray[np.complex_]:
    """
    ensures the coefficients is bandlimited for a given region
    """
    field = ssht.inverse(flm, L, Reality=reality, Spin=spin, Method=SAMPLING_SCHEME)
    mask = create_mask_region(L, region)
    field = np.where(mask, field, 0)
    return ssht.forward(field, L, Reality=reality, Spin=spin, Method=SAMPLING_SCHEME)


def create_default_region() -> Region:
    """
    creates default region
    """
    return Region(
        gap=POLAR_GAP,
        mask_name=SLEPIAN_MASK,
        phi_max=np.deg2rad(PHI_MAX),
        phi_min=np.deg2rad(PHI_MIN),
        theta_max=np.deg2rad(THETA_MAX),
        theta_min=np.deg2rad(THETA_MIN),
    )


def create_mesh_region(
    mesh_config: dict, vertices: npt.NDArray[np.float_]
) -> npt.NDArray[np.bool_]:
    """
    creates the boolean region for the given mesh
    """
    return (
        (vertices[:, 0] >= mesh_config["XMIN"])
        & (vertices[:, 0] <= mesh_config["XMAX"])
        & (vertices[:, 1] >= mesh_config["YMIN"])
        & (vertices[:, 1] <= mesh_config["YMAX"])
        & (vertices[:, 2] >= mesh_config["ZMIN"])
        & (vertices[:, 2] <= mesh_config["ZMAX"])
    )


def ensure_masked_bandlimit_mesh_signal(
    mesh: Mesh, u_i: npt.NDArray[np.complex_ | np.float_]
) -> npt.NDArray[np.float_]:
    """
    ensures that signal in pixel space is bandlimited
    """
    field = mesh_inverse(mesh, u_i)
    masked_field = np.where(mesh.region, field, 0)
    return mesh_forward(mesh, masked_field)


def convert_region_on_vertices_to_faces(mesh: Mesh) -> npt.NDArray[np.float_]:
    """
    converts the region on vertices to faces
    """
    region_reshape = np.argwhere(mesh.region).reshape(-1)
    faces_in_region = np.isin(mesh.faces, region_reshape).all(axis=1)
    region_on_faces = np.zeros(mesh.faces.shape[0])
    region_on_faces[faces_in_region] = 1
    return region_on_faces


def _create_africa_mask(
    L: int, earth_flm: npt.NDArray[np.complex_]
) -> npt.NDArray[np.float_]:
    """
    creates the Africa region mask
    """
    rot_flm = rotate_earth_to_africa(earth_flm, L)
    earth_f = ssht.inverse(rot_flm, L, Reality=True, Method=SAMPLING_SCHEME)
    thetas, _ = ssht.sample_positions(L, Grid=True, Method=SAMPLING_SCHEME)
    return (thetas <= AFRICA_RANGE) & (earth_f >= 0)


def _create_south_america_mask(
    L: int, earth_flm: npt.NDArray[np.complex_]
) -> npt.NDArray[np.float_]:
    """
    creates the Africa region mask
    """
    rot_flm = rotate_earth_to_south_america(earth_flm, L)
    earth_f = ssht.inverse(rot_flm, L, Reality=True, Method=SAMPLING_SCHEME)
    thetas, _ = ssht.sample_positions(L, Grid=True, Method=SAMPLING_SCHEME)
    return (thetas <= SOUTH_AMERICA_RANGE) & (earth_f >= 0)


def create_mask(L: int, mask_name: str) -> npt.NDArray[np.float_]:
    """
    creates the South America region mask
    """
    earth_flm = create_flm(L)
    if mask_name == f"africa_L{L}.npy":
        mask = _create_africa_mask(L, earth_flm)
    elif mask_name == f"south_america_L{L}.npy":
        mask = _create_south_america_mask(L, earth_flm)
    else:
        raise ValueError(f"Mask name {mask_name} not recognised")
    np.save(_data_path / f"slepian_masks_{mask_name}", mask)
    return mask



from dataclasses import KW_ONLY

from pydantic.dataclasses import dataclass

from sleplet.utils.mesh_methods import (
    create_mesh_region,
    extract_mesh_config,
    mesh_eigendecomposition,
    read_mesh,
)
from sleplet.utils.plotly_methods import create_camera
from sleplet.utils.validation import Validation


@dataclass(config=Validation)
class Mesh:
    name: str
    _: KW_ONLY
    number_basis_functions: int | None = None
    zoom: bool = False

    def __post_init_post_parse__(self) -> None:
        mesh_config = extract_mesh_config(self.name)
        self.camera_view = create_camera(
            mesh_config["CAMERA_X"],
            mesh_config["CAMERA_Y"],
            mesh_config["CAMERA_Z"],
            mesh_config["REGION_ZOOM"] if self.zoom else mesh_config["DEFAULT_ZOOM"],
            x_center=mesh_config["CENTER_X"] if self.zoom else 0,
            y_center=mesh_config["CENTER_Y"] if self.zoom else 0,
            z_center=mesh_config["CENTER_Z"] if self.zoom else 0,
        )
        self.colourbar_pos = (
            mesh_config["REGION_COLOURBAR_POS"]
            if self.zoom
            else mesh_config["DEFAULT_COLOURBAR_POS"]
        )
        self.vertices, self.faces = read_mesh(mesh_config)
        self.region = create_mesh_region(mesh_config, self.vertices)
        (
            self.mesh_eigenvalues,
            self.basis_functions,
            self.number_basis_functions,
        ) = mesh_eigendecomposition(
            self.name,
            self.vertices,
            self.faces,
            number_basis_functions=self.number_basis_functions,
        )



import numpy as np
import pytest
from numpy.testing import assert_allclose

from sleplet.utils.harmonic_methods import mesh_forward, mesh_inverse
from sleplet.utils.integration_methods import integrate_whole_mesh


def test_inverse_forward_transform_recovery(mesh_field_region) -> None:
    """
    tests that a given function is recovered after an
    inverse and forward transform on the mesh
    """
    u = mesh_inverse(mesh_field_region.mesh, mesh_field_region.coefficients)
    kernel_recov = mesh_forward(mesh_field_region.mesh, u)
    assert_allclose(
        np.abs(mesh_field_region.coefficients - kernel_recov).mean(), 0, atol=1e-14
    )


@pytest.mark.slow
def test_orthonormality_over_mesh_full(mesh) -> None:
    """
    for the computation of the Slepian D matrix the basis
    functions must be orthornomal over the whole mesh
    """
    orthonormality = np.zeros(
        (mesh.mesh_eigenvalues.shape[0], mesh.mesh_eigenvalues.shape[0])
    )
    for i, phi_i in enumerate(mesh.basis_functions):
        for j, phi_j in enumerate(mesh.basis_functions):
            orthonormality[i, j] = integrate_whole_mesh(
                mesh.vertices, mesh.faces, phi_i, phi_j
            )
    identity = np.identity(mesh.mesh_eigenvalues.shape[0])
    np.testing.assert_allclose(np.abs(orthonormality - identity).mean(), 0, atol=1e-16)



import numpy as np
from numpy.random import default_rng

from sleplet import logger
from sleplet.functions.fp.slepian_wavelets import SlepianWavelets
from sleplet.plotting.create_plot_sphere import Plot
from sleplet.utils.harmonic_methods import compute_random_signal
from sleplet.utils.region import Region
from sleplet.utils.slepian_methods import slepian_inverse
from sleplet.utils.vars import RANDOM_SEED
from sleplet.utils.wavelet_methods import (
    compute_slepian_wavelet_covariance,
    find_non_zero_wavelet_coefficients,
    slepian_wavelet_forward,
)

B = 3
J_MIN = 2
L = 128
NORMALISE = False
RUNS = 10
VAR_FP = 1


def main() -> None:
    """
    plots the difference between the theoretical &
    experimental covariances for the Slepian wavelets
    """
    # compute wavelets
    region = Region(mask_name="africa")
    sw = SlepianWavelets(L, B=B, j_min=J_MIN, region=region)

    # theoretical covariance
    covar_theory = compute_slepian_wavelet_covariance(
        L, sw.wavelets, sw.slepian, var_signal=VAR_FP
    )

    # initialise matrix
    covar_runs_shape = (RUNS, *covar_theory.shape)
    covar_data_runs = np.zeros(covar_runs_shape, dtype=np.complex_)

    # set seed
    rng = default_rng(RANDOM_SEED)

    for i in range(RUNS):
        # Generate normally distributed random complex signal
        f_p = compute_random_signal(L, rng, var_signal=VAR_FP)

        # compute wavelet coefficients
        w_p = slepian_wavelet_forward(f_p, sw.wavelets, sw.slepian.N)

        # compute field values
        for j, coefficient in enumerate(w_p):
            logger.info(f"run: {i+1}/{RUNS}, compute covariance: {j+1}/{len(w_p)}")
            covar_data_runs[i, j] = slepian_inverse(coefficient, L, sw.slepian)

    # compute covariance
    covar_data = covar_data_runs.var(axis=0)

    # compute differences
    non_zero_theory = find_non_zero_wavelet_coefficients(covar_theory, axis=(1, 2))
    non_zero_data = find_non_zero_wavelet_coefficients(covar_data, axis=(1, 2))
    differences = np.abs(non_zero_data - non_zero_theory) / non_zero_theory

    # report errors
    for j, diff in enumerate(differences):
        name = f"slepian_covariance_diff_{j}"
        logger.info(name)
        Plot(diff, L, name, normalise=NORMALISE, region=sw.region).execute()


if __name__ == "__main__":
    main()



import numpy as np

from sleplet import logger
from sleplet.functions.fp.slepian_wavelet_coefficients_south_america import (
    SlepianWaveletCoefficientsSouthAmerica,
)
from sleplet.plotting.create_plot_sphere import Plot
from sleplet.utils.region import Region
from sleplet.utils.slepian_methods import slepian_inverse
from sleplet.utils.vars import SMOOTHING
from sleplet.utils.wavelet_methods import slepian_wavelet_inverse

B = 3
J_MIN = 2
L = 128
NORMALISE = False


def main() -> None:
    """
    the reconstruction of a signal in Slepian space
    """
    region = Region(mask_name="south_america")
    swc = SlepianWaveletCoefficientsSouthAmerica(
        L, B=B, j_min=J_MIN, region=region, smoothing=SMOOTHING
    )

    # plot
    f_p = np.zeros(swc.slepian.N, dtype=np.complex_)
    for p, coeff in enumerate(swc.wavelet_coefficients):
        logger.info(f"plot reconstruction: {p}")
        f_p += slepian_wavelet_inverse(coeff, swc.wavelets, swc.slepian.N)
        f = slepian_inverse(f_p, L, swc.slepian)
        name = f"south_america_wavelet_reconstruction_progressive_{p}_L{L}"
        Plot(f, L, name, normalise=NORMALISE, region=swc.region).execute()


if __name__ == "__main__":
    main()



import numpy as np
from numpy.testing import assert_allclose

from sleplet.utils.harmonic_methods import mesh_inverse
from sleplet.utils.slepian_methods import slepian_mesh_forward, slepian_mesh_inverse
from sleplet.utils.wavelet_methods import (
    slepian_wavelet_forward,
    slepian_wavelet_inverse,
)


def test_decompose_all_mesh(mesh_slepian, mesh_field_region) -> None:
    """
    tests that all three methods produce the same coefficients for the mesh
    """
    field = mesh_inverse(mesh_slepian.mesh, mesh_field_region.coefficients)
    harmonic_sum_p = slepian_mesh_forward(
        mesh_slepian,
        u_i=mesh_field_region.coefficients,
    )
    integrate_sphere_p = slepian_mesh_forward(
        mesh_slepian,
        u=field,
    )
    integrate_region_p = slepian_mesh_forward(
        mesh_slepian,
        u=field,
        mask=True,
    )
    assert_allclose(
        np.abs(integrate_sphere_p - harmonic_sum_p)[: mesh_slepian.N].mean(),
        0,
        atol=1e-14,
    )
    assert_allclose(
        np.abs(integrate_region_p - harmonic_sum_p)[: mesh_slepian.N].mean(),
        0,
        atol=1e-14,
    )


def test_forward_inverse_transform_slepian(mesh_slepian, mesh_field_region) -> None:
    """
    tests that the Slepian forward and inverse transforms recover the field
    """
    f_p = slepian_mesh_forward(
        mesh_slepian,
        u_i=mesh_field_region.coefficients,
    )
    f_slepian = slepian_mesh_inverse(mesh_slepian, f_p)
    f_harmonic = mesh_inverse(mesh_slepian.mesh, mesh_field_region.coefficients)
    assert_allclose(
        np.abs(f_slepian - f_harmonic)[mesh_slepian.mesh.region].mean(),
        0,
        atol=7e-3,
    )


def test_synthesis_mesh(mesh_slepian_wavelets, mesh_field_region) -> None:
    """
    tests that Slepian polar wavelet synthesis matches the coefficients
    """
    coefficients = slepian_mesh_forward(
        mesh_slepian_wavelets.mesh_slepian,
        u_i=mesh_field_region.coefficients,
    )
    wav_coeffs = slepian_wavelet_forward(
        coefficients,
        mesh_slepian_wavelets.wavelets,
        mesh_slepian_wavelets.mesh_slepian.N,
    )
    f_p = slepian_wavelet_inverse(
        wav_coeffs,
        mesh_slepian_wavelets.wavelets,
        mesh_slepian_wavelets.mesh_slepian.N,
    )
    assert_allclose(
        np.abs(f_p - coefficients)[: mesh_slepian_wavelets.mesh_slepian.N].mean(),
        0,
        atol=1e-16,
    )



from argparse import ArgumentParser

from sleplet import logger
from sleplet.functions.fp.slepian_africa import SlepianAfrica
from sleplet.functions.fp.slepian_wavelets import SlepianWavelets
from sleplet.plotting.create_plot_sphere import Plot
from sleplet.utils.denoising import denoising_slepian_wavelet
from sleplet.utils.plot_methods import find_max_amplitude
from sleplet.utils.region import Region
from sleplet.utils.string_methods import filename_args
from sleplet.utils.vars import SMOOTHING

B = 3
J_MIN = 2
L = 128
N_SIGMA = 2
NORMALISE = False
SNR_IN = -10


def main(snr: float, sigma: int) -> None:
    """
    denoising demo using Slepian wavelets
    """
    logger.info(f"SNR={snr}, n_sigma={sigma}")
    # setup
    region = Region(mask_name="africa")

    # create map & noised map
    fun = SlepianAfrica(L, region=region, smoothing=SMOOTHING)
    fun_noised = SlepianAfrica(L, noise=snr, region=region, smoothing=SMOOTHING)

    # create wavelets
    sw = SlepianWavelets(L, B=B, j_min=J_MIN, region=region)

    # fix amplitude
    amplitude = find_max_amplitude(fun)

    f = denoising_slepian_wavelet(fun, fun_noised, sw, snr, sigma)
    name = f"{fun.name}{filename_args(snr, 'snr')}{filename_args(sigma,'n')}_denoised"
    Plot(
        f, L, name, amplitude=amplitude, normalise=NORMALISE, region=sw.region
    ).execute()


if __name__ == "__main__":
    parser = ArgumentParser(description="denoising")
    parser.add_argument(
        "--noise",
        "-n",
        type=int,
        default=SNR_IN,
    )
    parser.add_argument(
        "--sigma",
        "-s",
        type=int,
        default=N_SIGMA,
    )
    args = parser.parse_args()
    main(args.noise, args.sigma)



from glob import glob
from pathlib import Path

from sleplet.functions.coefficients import Coefficients
from sleplet.functions.flm.africa import Africa
from sleplet.functions.flm.axisymmetric_wavelet_coefficients_africa import (
    AxisymmetricWaveletCoefficientsAfrica,
)
from sleplet.functions.flm.axisymmetric_wavelet_coefficients_earth import (
    AxisymmetricWaveletCoefficientsEarth,
)
from sleplet.functions.flm.axisymmetric_wavelet_coefficients_south_america import (
    AxisymmetricWaveletCoefficientsSouthAmerica,
)
from sleplet.functions.flm.axisymmetric_wavelets import AxisymmetricWavelets
from sleplet.functions.flm.dirac_delta import DiracDelta
from sleplet.functions.flm.directional_spin_wavelets import DirectionalSpinWavelets
from sleplet.functions.flm.earth import Earth
from sleplet.functions.flm.elongated_gaussian import ElongatedGaussian
from sleplet.functions.flm.gaussian import Gaussian
from sleplet.functions.flm.harmonic_gaussian import HarmonicGaussian
from sleplet.functions.flm.identity import Identity
from sleplet.functions.flm.noise_earth import NoiseEarth
from sleplet.functions.flm.ridgelets import Ridgelets
from sleplet.functions.flm.south_america import SouthAmerica
from sleplet.functions.flm.spherical_harmonic import SphericalHarmonic
from sleplet.functions.flm.squashed_gaussian import SquashedGaussian
from sleplet.functions.flm.wmap import Wmap
from sleplet.functions.fp.slepian import Slepian
from sleplet.functions.fp.slepian_africa import SlepianAfrica
from sleplet.functions.fp.slepian_dirac_delta import SlepianDiracDelta
from sleplet.functions.fp.slepian_identity import SlepianIdentity
from sleplet.functions.fp.slepian_noise_africa import SlepianNoiseAfrica
from sleplet.functions.fp.slepian_noise_south_america import SlepianNoiseSouthAmerica
from sleplet.functions.fp.slepian_south_america import SlepianSouthAmerica
from sleplet.functions.fp.slepian_wavelet_coefficients_africa import (
    SlepianWaveletCoefficientsAfrica,
)
from sleplet.functions.fp.slepian_wavelet_coefficients_south_america import (
    SlepianWaveletCoefficientsSouthAmerica,
)
from sleplet.functions.fp.slepian_wavelets import SlepianWavelets
from sleplet.meshes.harmonic_coefficients.mesh_basis_functions import MeshBasisFunctions
from sleplet.meshes.harmonic_coefficients.mesh_field import MeshField
from sleplet.meshes.harmonic_coefficients.mesh_noise_field import MeshNoiseField
from sleplet.meshes.mesh_coefficients import MeshCoefficients
from sleplet.meshes.slepian_coefficients.mesh_slepian_field import MeshSlepianField
from sleplet.meshes.slepian_coefficients.mesh_slepian_functions import (
    MeshSlepianFunctions,
)
from sleplet.meshes.slepian_coefficients.mesh_slepian_noise_field import (
    MeshSlepianNoiseField,
)
from sleplet.meshes.slepian_coefficients.mesh_slepian_wavelet_coefficients import (
    MeshSlepianWaveletCoefficients,
)
from sleplet.meshes.slepian_coefficients.mesh_slepian_wavelets import (
    MeshSlepianWavelets,
)

_data_path = Path(__file__).resolve().parents[1] / "data"
MESHES: list[str] = [
    Path(x).stem.removeprefix("meshes_regions_")
    for x in glob(str(_data_path / "*.toml"))
]

FLM: list[type[Coefficients]] = [
    Africa,
    AxisymmetricWaveletCoefficientsAfrica,
    AxisymmetricWaveletCoefficientsEarth,
    AxisymmetricWaveletCoefficientsSouthAmerica,
    AxisymmetricWavelets,
    DiracDelta,
    DirectionalSpinWavelets,
    Earth,
    ElongatedGaussian,
    Gaussian,
    HarmonicGaussian,
    Identity,
    NoiseEarth,
    Ridgelets,
    SouthAmerica,
    SphericalHarmonic,
    SquashedGaussian,
    Wmap,
]

FP: list[type[Coefficients]] = [
    Slepian,
    SlepianAfrica,
    SlepianDiracDelta,
    SlepianIdentity,
    SlepianNoiseAfrica,
    SlepianNoiseSouthAmerica,
    SlepianSouthAmerica,
    SlepianWaveletCoefficientsAfrica,
    SlepianWaveletCoefficientsSouthAmerica,
    SlepianWavelets,
]

COEFFICIENTS: list[type[Coefficients]] = FLM + FP

MAPS_LM: list[type[Coefficients]] = [Earth, SouthAmerica, Wmap]

MESH_HARMONIC: list[type[MeshCoefficients]] = [
    MeshBasisFunctions,
    MeshField,
    MeshNoiseField,
]

MESH_SLEPIAN: list[type[MeshCoefficients]] = [
    MeshSlepianField,
    MeshSlepianFunctions,
    MeshSlepianNoiseField,
    MeshSlepianWaveletCoefficients,
    MeshSlepianWavelets,
]

MESH_COEFFICIENTS: list[type[MeshCoefficients]] = MESH_HARMONIC + MESH_SLEPIAN



from functools import reduce
from typing import Any

import numpy as np
import pyssht as ssht
from numpy import typing as npt

from sleplet.utils.vars import SAMPLING_SCHEME


def calc_integration_weight(L: int) -> npt.NDArray[np.float_]:
    """
    computes the spherical Jacobian for the integration
    """
    thetas, phis = ssht.sample_positions(L, Grid=True, Method=SAMPLING_SCHEME)
    delta_theta = np.ediff1d(thetas[:, 0]).mean()
    delta_phi = np.ediff1d(phis[0]).mean()
    return np.sin(thetas) * delta_theta * delta_phi


def integrate_whole_sphere(
    weight: npt.NDArray[np.float_], *functions: npt.NDArray[np.complex_]
) -> complex:
    """
    computes the integration for the whole sphere
    """
    multiplied_inputs = _multiply_args(*functions)
    return (multiplied_inputs * weight).sum()


def integrate_region_sphere(
    mask: npt.NDArray[np.float_],
    weight: npt.NDArray[np.float_],
    *functions: npt.NDArray[np.complex_ | np.float_],
) -> complex:
    """
    computes the integration for a region of the sphere
    """
    multiplied_inputs = _multiply_args(*functions)
    return (multiplied_inputs * weight * mask).sum()


def integrate_whole_mesh(
    vertices: npt.NDArray[np.float_],
    faces: npt.NDArray[np.int_],
    *functions: npt.NDArray[np.complex_ | np.float_],
) -> float:
    """
    computes the integral of functions on the vertices
    """
    multiplied_inputs = _multiply_args(*functions)
    return multiplied_inputs.sum()


def integrate_region_mesh(
    mask: npt.NDArray[np.bool_],
    vertices: npt.NDArray[np.float_],
    faces: npt.NDArray[np.int_],
    *functions: npt.NDArray[np.complex_ | np.float_],
) -> float:
    """
    computes the integral of a region of functions on the vertices
    """
    multiplied_inputs = _multiply_args(*functions)
    return (multiplied_inputs * mask).sum()


def _multiply_args(*args: npt.NDArray[Any]) -> npt.NDArray[Any]:
    """
    method to multiply an unknown number of arguments
    """
    return reduce((lambda x, y: x * y), args)



from pathlib import Path

import numpy as np
import pandas as pd
import scipy.io as sio

_data_path = Path(__file__).resolve().parents[3] / "data"


def _add_missing_data(df: pd.DataFrame) -> pd.DataFrame:
    """
    the tidal data doesn't have value for ell=0 or ell=1
    """
    df_missing = pd.DataFrame(
        [[0, 0, 0, 0], [1, 0, 0, 0], [1, 0, 0, 0]],
        columns=df.columns,
    )
    return df.append(df_missing, ignore_index=True)


def _helper(df: pd.DataFrame, filename: str) -> None:
    """
    method which gets the dataframe in the correct format and saves as mat file
    """
    # swap 'D' with 'E' in floats to convert to numeric
    df["real"] = pd.to_numeric(df["real"].str.replace("D", "E"))
    df["imag"] = pd.to_numeric(df["imag"].str.replace("D", "E"))

    # add missing tidal data
    df = df if (df["ell"] == 0).any() else _add_missing_data(df)

    # bandlimit
    L = df["ell"].max()

    # fill dataframe with zero for missing values
    # the original data doesn't have negative indices
    lst: list = []
    for ell in range(L + 1):
        lst.extend([ell, m, 0, 0] for m in range(-ell, 0))
    df_negatives = pd.DataFrame(np.array(lst), columns=df.columns.to_list())

    # combine dataframes and sort on ell & m
    df_concat = pd.concat([df, df_negatives])
    df_sorted = df_concat.sort_values(["ell", "m"]).reset_index(drop=True)

    # make complex flms
    df_sorted["flm"] = df_sorted["real"] + 1j * df_sorted["imag"]

    filename = f"EGM2008_{filename}_L{L}.mat"
    flm = np.array(df_sorted["flm"])
    mat = {"L": L, "flm": flm}
    sio.savemat(_data_path / filename, mat, oned_as="column")


def create_matfile(filename: str) -> None:
    """
    read original data file from
    Spherical Harmonic Coefficients for Earth's Elevation
    http://earth-info.nga.mil/GandG/wgs84/gravitymod/egm2008/first_release.html
    """
    # filename inputs and outputs
    inputs = [
        "Coeff_Height_and_Depth_to2190_DTM2006.0",
        "EGM2008_to2190_TideFree",
        "EGM2008_to2190_ZeroTide",
    ]
    outputs = ["Topography_flms", "tide_free", "zero_tide"]

    # read the data and drop error columns
    col_names = ["ell", "m", "real", "imag", "error_real", "error_imag"]
    df = pd.read_csv(_data_path / filename, delimiter="\\s+", names=col_names)
    df.drop(columns=["error_real", "error_imag"], inplace=True)

    output = outputs[inputs.index(filename)]
    _helper(df, output)


if __name__ == "__main__":
    filename = "Coeff_Height_and_Depth_to2190_DTM2006.0"
    # filename = "EGM2008_to2190_TideFree"
    # filename = "EGM2008_to2190_ZeroTide"
    create_matfile(filename)



from pathlib import Path

import numpy as np
import seaborn as sns
from matplotlib import pyplot as plt
from numpy import typing as npt

from sleplet import logger
from sleplet.slepian.slepian_region.slepian_polar_cap import SlepianPolarCap
from sleplet.utils.harmonic_methods import invert_flm_boosted
from sleplet.utils.plot_methods import calc_plot_resolution, save_plot
from sleplet.utils.vars import THETA_MAX_DEFAULT, THETA_MIN_DEFAULT

L = 16
ORDERS = 4
PHI_IDX = 0
RANKS = 3
RESOLUTION = calc_plot_resolution(L)
SIGNS = [[1, -1, 1], [-1, -1, -1], [1, 1, 1], [1, 1, -1]]
TEXT_BOX: dict[str, str | float] = {"boxstyle": "round", "color": "w"}
THETA_MAX = 40


fig_path = Path(__file__).resolve().parents[2] / "src" / "sleplet" / "figures"
sns.set(context="paper")


def main() -> None:
    """
    create fig 5.1 from Spatiospectral Concentration on a Sphere by Simons et al 2006
    """
    x = np.linspace(THETA_MIN_DEFAULT, np.rad2deg(THETA_MAX_DEFAULT), RESOLUTION + 1)
    i = (x < THETA_MAX).sum()
    _, ax = plt.subplots(ORDERS, RANKS, sharex="col", sharey="row")
    plt.setp(
        ax,
        xlim=[0, 180],
        xticks=[0, 40, 180],
        xticklabels=[r"$0^\circ$", r"$40^\circ$", r"$180^\circ$"],
        ylim=[-3, 3],
        yticks=[-2, 0, 2],
    )
    for order in range(ORDERS):
        slepian = SlepianPolarCap(L, np.deg2rad(THETA_MAX), order=order)
        for rank in range(RANKS):
            _helper(ax, slepian, RESOLUTION, x, i, order, rank)
    save_plot(fig_path, "slepian_colatitude")


def _helper(  # noqa: PLR0913
    ax: npt.NDArray,
    slepian: SlepianPolarCap,
    resolution: int,
    x: npt.NDArray[np.float_],
    i: int,
    order: int,
    rank: int,
) -> None:
    """
    helper which plots the required order and specified ranks
    """
    logger.info(f"plotting order={order}, rank={rank}")
    axs = ax[order, rank]
    flm = slepian.eigenvectors[rank] * SIGNS[order][rank]
    lam = slepian.eigenvalues[rank]
    f = invert_flm_boosted(flm, L, resolution).real
    if rank == 0:
        axs.set_ylabel(rf"$m={{{order}}}$")
    if order == 0:
        axs.set_title(rf"$\alpha={{{rank+1}}}$")
    if order == ORDERS - 1:
        axs.set_xlabel("colatitude")
    axs.plot(x[:i], f[:i, PHI_IDX], x[i:], f[i:, PHI_IDX])
    axs.text(
        0.45, 0.81, rf"$\mu={{{lam:.6f}}}$", transform=axs.transAxes, bbox=TEXT_BOX
    )


if __name__ == "__main__":
    main()



import numpy as np
from numpy import typing as npt
from pydantic import validator
from pydantic.dataclasses import dataclass
from pys2let import pys2let_j_max

from sleplet import logger
from sleplet.functions.f_lm import F_LM
from sleplet.functions.flm.earth import Earth
from sleplet.utils.string_methods import (
    convert_camel_case_to_snake_case,
    filename_args,
    wavelet_ending,
)
from sleplet.utils.validation import Validation
from sleplet.utils.wavelet_methods import (
    axisymmetric_wavelet_forward,
    create_axisymmetric_wavelets,
)


@dataclass(config=Validation, kw_only=True)
class AxisymmetricWaveletCoefficientsEarth(F_LM):
    B: int = 3
    j_min: int = 2
    j: int | None = None

    def __post_init_post_parse__(self) -> None:
        super().__post_init_post_parse__()

    def _create_coefficients(self) -> npt.NDArray[np.complex_ | np.float_]:
        logger.info("start computing wavelet coefficients")
        self.wavelets, self.wavelet_coefficients = self._create_wavelet_coefficients()
        logger.info("finish computing wavelet coefficients")
        jth = 0 if self.j is None else self.j + 1
        return self.wavelet_coefficients[jth]

    def _create_name(self) -> str:
        return (
            f"{convert_camel_case_to_snake_case(self.__class__.__name__)}"
            f"{filename_args(self.B, 'B')}"
            f"{filename_args(self.j_min, 'jmin')}"
            f"{wavelet_ending(self.j_min, self.j)}"
        )

    def _set_reality(self) -> bool:
        return False

    def _set_spin(self) -> int:
        return 0

    def _setup_args(self) -> None:
        if isinstance(self.extra_args, list):
            num_args = 3
            if len(self.extra_args) != num_args:
                raise ValueError(f"The number of extra arguments should be {num_args}")
            self.B, self.j_min, self.j = self.extra_args

    def _create_wavelet_coefficients(
        self,
    ) -> tuple[npt.NDArray[np.complex_], npt.NDArray[np.complex_]]:
        """
        computes wavelet coefficients of the Earth
        """
        wavelets = create_axisymmetric_wavelets(self.L, self.B, self.j_min)
        self.earth = Earth(self.L, smoothing=self.smoothing)
        wavelet_coefficients = axisymmetric_wavelet_forward(
            self.L, self.earth.coefficients, wavelets
        )
        return wavelets, wavelet_coefficients

    @validator("j")
    def check_j(cls, v, values):
        j_max = pys2let_j_max(values["B"], values["L"], values["j_min"])
        if v is not None and v < 0:
            raise ValueError("j should be positive")
        if v is not None and v > j_max - values["j_min"]:
            raise ValueError(
                f"j should be less than j_max - j_min: {j_max - values['j_min'] + 1}"
            )
        return v



import numpy as np
from numpy import typing as npt
from pydantic import validator
from pydantic.dataclasses import dataclass
from pys2let import pys2let_j_max

from sleplet import logger
from sleplet.meshes.mesh_slepian_coefficients import MeshSlepianCoefficients
from sleplet.utils.string_methods import filename_args, wavelet_ending
from sleplet.utils.validation import Validation
from sleplet.utils.wavelet_methods import create_kappas


@dataclass(config=Validation, kw_only=True)
class MeshSlepianWavelets(MeshSlepianCoefficients):
    B: int = 3
    j_min: int = 2
    j: int | None = None

    def __post_init_post_parse__(self) -> None:
        super().__post_init_post_parse__()

    def _create_coefficients(self) -> npt.NDArray[np.complex_ | np.float_]:
        logger.info("start computing wavelets")
        self.wavelets = self._create_wavelets()
        logger.info("finish computing wavelets")
        jth = 0 if self.j is None else self.j + 1
        return self.wavelets[jth]

    def _create_name(self) -> str:
        return (
            f"slepian_wavelets_{self.mesh.name}"
            f"{filename_args(self.B, 'B')}"
            f"{filename_args(self.j_min, 'jmin')}"
            f"{wavelet_ending(self.j_min, self.j)}"
        )

    def _setup_args(self) -> None:
        if isinstance(self.extra_args, list):
            num_args = 3
            if len(self.extra_args) != num_args:
                raise ValueError(f"The number of extra arguments should be {num_args}")
            self.B, self.j_min, self.j = self.extra_args

    def _create_wavelets(self) -> npt.NDArray[np.float_]:
        """
        creates the Slepian wavelets of the mesh
        """
        return create_kappas(self.mesh.mesh_eigenvalues.shape[0], self.B, self.j_min)

    @validator("j")
    def check_j(cls, v, values) -> int | None:
        j_max = pys2let_j_max(
            values["B"], values["mesh"].mesh_eigenvalues.shape[0], values["j_min"]
        )
        if v is not None and v < 0:
            raise ValueError("j should be positive")
        if v is not None and v > j_max - values["j_min"]:
            raise ValueError(
                f"j should be less than j_max - j_min: {j_max - values['j_min'] + 1}"
            )
        return v



from pathlib import Path

import numpy as np
import pyssht as ssht
import seaborn as sns
from matplotlib import pyplot as plt
from numpy import typing as npt

from sleplet.functions.flm.earth import Earth
from sleplet.slepian.slepian_functions import SlepianFunctions
from sleplet.slepian.slepian_region.slepian_polar_cap import SlepianPolarCap
from sleplet.utils.plot_methods import save_plot
from sleplet.utils.region import Region
from sleplet.utils.slepian_methods import choose_slepian_method, slepian_forward
from sleplet.utils.vars import SAMPLING_SCHEME

L = 16
THETA_MAX = 40

fig_path = Path(__file__).resolve().parents[2] / "src" / "sleplet" / "figures"
sns.set(context="paper")


def main() -> None:
    """
    creates a plot of Slepian coefficients against rank
    """
    region = Region(theta_max=np.deg2rad(THETA_MAX))
    earth = Earth(L, region=region)
    slepian = choose_slepian_method(L, region)
    field = ssht.inverse(earth.coefficients, L, Method=SAMPLING_SCHEME)
    integrate_region = _helper_region(
        L, slepian, field, earth.coefficients, slepian.mask
    )
    integrate_sphere = _helper_sphere(L, slepian, field, earth.coefficients)
    N = SlepianPolarCap(L, np.deg2rad(THETA_MAX)).N
    ax = plt.gca()
    sns.scatterplot(
        x=range(N), y=integrate_region, ax=ax, label="region", linewidth=0, marker="."
    )
    sns.scatterplot(
        x=range(N), y=integrate_sphere, ax=ax, label="sphere", linewidth=0, marker="*"
    )
    ax.set_xlabel("coefficients")
    ax.set_ylabel("relative error")
    ax.set_yscale("log")
    save_plot(fig_path, f"fp_error_earth_polar{THETA_MAX}_L{L}")


def _helper_sphere(
    L: int,
    slepian: SlepianFunctions,
    f: npt.NDArray[np.complex_],
    flm: npt.NDArray[np.complex_ | np.float_],
) -> npt.NDArray[np.float_]:
    """
    the difference in Slepian coefficients by integration of whole sphere
    """
    output = np.abs(slepian_forward(L, slepian, f=f))
    desired = np.abs(slepian_forward(L, slepian, flm=flm))
    return np.abs(output - desired) / desired


def _helper_region(
    L: int,
    slepian: SlepianFunctions,
    f: npt.NDArray[np.complex_],
    flm: npt.NDArray[np.complex_ | np.float_],
    mask: npt.NDArray[np.float_],
) -> npt.NDArray[np.float_]:
    """
    the difference in Slepian coefficients by integration of region on the sphere
    """
    output = np.abs(slepian_forward(L, slepian, f=f, mask=mask))
    desired = np.abs(slepian_forward(L, slepian, flm=flm))
    return np.abs(output - desired) / desired


if __name__ == "__main__":
    main()



import cmocean
import numpy as np
from numpy.testing import assert_equal

from sleplet.utils.plot_methods import (
    calc_nearest_grid_point,
    calc_plot_resolution,
    convert_colourscale,
)

L = 128
PHI_0 = np.pi / 6
PL_ENTRIES = 255
THETA_MAX = np.pi / 3


def test_resolution_values() -> None:
    """
    verifies the correct resolution is chosen for a given bandlimit
    """
    arguments = [1, 10, 100, 1000]
    output = [64, 80, 800, 2000]
    for c, arg in enumerate(arguments):
        assert_equal(calc_plot_resolution(arg), output[c])


def test_create_colourscale() -> None:
    """
    test creates a plotly compatible colourscale
    """
    colourscale = convert_colourscale(cmocean.cm.ice, pl_entries=PL_ENTRIES)
    assert_equal(len(colourscale), PL_ENTRIES)


def test_find_nearest_grid_point() -> None:
    """
    test to find nearest grid point to provided angles
    """
    alpha, beta = calc_nearest_grid_point(L, PHI_0 / np.pi, THETA_MAX / np.pi)
    assert_equal(alpha, 0.5154175447295755)
    assert_equal(beta, 1.055378782065321)



#!/usr/bin/env python
from argparse import ArgumentParser, Namespace

import numpy as np
from numpy import typing as npt

from sleplet import logger
from sleplet.meshes.classes.mesh import Mesh
from sleplet.meshes.mesh_coefficients import MeshCoefficients
from sleplet.meshes.mesh_slepian_coefficients import MeshSlepianCoefficients
from sleplet.plotting.create_plot_mesh import Plot
from sleplet.utils.class_lists import MESH_COEFFICIENTS, MESHES
from sleplet.utils.harmonic_methods import mesh_inverse
from sleplet.utils.slepian_methods import slepian_mesh_inverse
from sleplet.utils.string_methods import convert_classes_list_to_snake_case


def valid_meshes(mesh_name: str) -> str:
    """
    check if valid mesh name
    """
    if mesh_name in MESHES:
        return mesh_name
    else:
        raise ValueError(f"'{mesh_name}' is not a valid mesh name to plot")


def valid_methods(method_name: str) -> str:
    """
    check if valid mesh name
    """
    if method_name in convert_classes_list_to_snake_case(
        MESH_COEFFICIENTS, word_to_remove="Mesh"
    ):
        return method_name
    else:
        raise ValueError(f"'{method_name}' is not a valid method to plot")


def read_args() -> Namespace:
    """
    method to read args from the command line
    """
    parser = ArgumentParser(description="Create mesh plot")
    parser.add_argument(
        "function",
        type=valid_meshes,
        choices=MESHES,
        help="mesh to plot",
    )
    parser.add_argument(
        "--extra_args",
        "-e",
        type=int,
        nargs="+",
        help="list of extra args for functions",
    )
    parser.add_argument(
        "--method",
        "-m",
        type=valid_methods,
        nargs="?",
        default="basis_functions",
        const="basis_functions",
        choices=convert_classes_list_to_snake_case(
            MESH_COEFFICIENTS, word_to_remove="Mesh"
        ),
        help="plotting routine: defaults to basis",
    )
    parser.add_argument("--noise", "-n", type=int, help="the SNR_IN of the noise level")
    parser.add_argument(
        "--region",
        "-r",
        action="store_true",
        help="flag which masks the function for a region",
    )
    parser.add_argument(
        "--unnormalise",
        "-u",
        action="store_true",
        help="flag turns off normalisation for plot",
    )
    parser.add_argument(
        "--zoom",
        "-z",
        action="store_true",
        help="flag which zooms in on the region of interest",
    )
    return parser.parse_args()


def plot(f: MeshCoefficients, *, normalise: bool, amplitude: float | None) -> None:
    """
    master plotting method
    """
    field = _coefficients_to_field(f, f.coefficients)
    Plot(
        f.mesh,
        f.name,
        field,
        amplitude=amplitude,
        normalise=normalise,
        region=isinstance(f, MeshSlepianCoefficients),
    ).execute()


def _coefficients_to_field(
    f: MeshCoefficients, coefficients: npt.NDArray[np.complex_ | np.float_]
) -> npt.NDArray[np.complex_ | np.float_]:
    """
    computes the field over the whole mesh from the harmonic/Slepian coefficients
    """
    return (
        slepian_mesh_inverse(f.mesh_slepian, coefficients)
        if isinstance(f, MeshSlepianCoefficients)
        else mesh_inverse(f.mesh, coefficients)
    )


def compute_amplitude_for_noisy_plots(f: MeshCoefficients) -> float | None:
    """
    for the noised plots fix the amplitude to the initial data
    """
    return (
        np.abs(_coefficients_to_field(f, f.unnoised_coefficients)).max()
        if f.unnoised_coefficients is not None
        else None
    )


def main() -> None:
    args = read_args()
    logger.info(f"mesh: '{args.function}', plotting method: '{args.method}'")

    # function to plot
    mesh = Mesh(args.function, zoom=args.zoom)
    f = MESH_COEFFICIENTS[
        convert_classes_list_to_snake_case(
            MESH_COEFFICIENTS, word_to_remove="Mesh"
        ).index(args.method)
    ](
        mesh,
        extra_args=args.extra_args,
        noise=args.noise if args.noise is not None else None,
        region=args.region,
    )

    # custom amplitude for noisy plots
    amplitude = compute_amplitude_for_noisy_plots(f)

    # perform plot
    plot(f, normalise=not args.unnormalise, amplitude=amplitude)


if __name__ == "__main__":
    main()



from concurrent.futures import ThreadPoolExecutor
from pathlib import Path

import numpy as np
from numpy import linalg as LA  # noqa: N812
from numpy import typing as npt
from pydantic.dataclasses import dataclass

from sleplet import logger
from sleplet.data.setup_pooch import find_on_pooch_then_local
from sleplet.meshes.classes.mesh import Mesh
from sleplet.utils.array_methods import fill_upper_triangle_of_hermitian_matrix
from sleplet.utils.integration_methods import integrate_region_mesh
from sleplet.utils.parallel_methods import (
    attach_to_shared_memory_block,
    create_shared_memory_array,
    free_shared_memory,
    release_shared_memory,
    split_arr_into_chunks,
)
from sleplet.utils.slepian_arbitrary_methods import compute_mesh_shannon
from sleplet.utils.validation import Validation
from sleplet.utils.vars import NCPU

_data_path = Path(__file__).resolve().parents[2] / "data"


@dataclass(config=Validation)
class MeshSlepian:
    mesh: Mesh

    def __post_init_post_parse__(self) -> None:
        self.N = compute_mesh_shannon(self.mesh)
        self._compute_slepian_functions()

    def _compute_slepian_functions(self) -> None:
        """
        computes the Slepian functions of the mesh
        """
        logger.info("computing slepian functions of mesh")

        # create filenames
        eigd_loc = (
            f"meshes_laplacians_slepian_functions_{self.mesh.name}_"
            f"b{self.mesh.mesh_eigenvalues.shape[0]}_N{self.N}"
        )
        eval_loc = f"{eigd_loc}_eigenvalues.npy"
        evec_loc = f"{eigd_loc}_eigenvectors.npy"

        try:
            self.slepian_eigenvalues = np.load(find_on_pooch_then_local(eval_loc))
            self.slepian_functions = np.load(find_on_pooch_then_local(evec_loc))
        except TypeError:
            self._compute_slepian_functions_from_scratch(eval_loc, evec_loc)

    def _compute_slepian_functions_from_scratch(self, eval_loc, evec_loc):
        D = self._create_D_matrix()
        logger.info(
            f"Shannon number from vertices: {self.N}, "
            f"Trace of D matrix: {round(D.trace())}, "
            f"difference: {round(np.abs(self.N - D.trace()))}"
        )

        # fill in remaining triangle section
        fill_upper_triangle_of_hermitian_matrix(D)

        # solve eigenproblem
        (
            self.slepian_eigenvalues,
            self.slepian_functions,
        ) = self._clean_evals_and_evecs(LA.eigh(D))
        np.save(_data_path / eval_loc, self.slepian_eigenvalues)
        np.save(_data_path / evec_loc, self.slepian_functions[: self.N])

    def _create_D_matrix(self) -> npt.NDArray[np.float_]:  # noqa: N802
        """
        computes the D matrix for the mesh eigenfunctions
        """
        D = np.zeros(
            (self.mesh.mesh_eigenvalues.shape[0], self.mesh.mesh_eigenvalues.shape[0])
        )

        D_ext, shm_ext = create_shared_memory_array(D)

        def func(chunk: list[int]) -> None:
            """
            calculate D matrix components for each chunk
            """
            D_int, shm_int = attach_to_shared_memory_block(D, shm_ext)

            for i in chunk:
                logger.info(f"start basis function: {i}")
                self._fill_D_elements(D_int, i)
                logger.info(f"finish basis function: {i}")

            free_shared_memory(shm_int)

        # split up L range to maximise effiency
        chunks = split_arr_into_chunks(self.mesh.mesh_eigenvalues.shape[0], NCPU)

        # initialise pool and apply function
        with ThreadPoolExecutor(max_workers=NCPU) as e:
            e.map(func, chunks)

        # retrieve from parallel function
        D = D_ext.copy()

        # Free and release the shared memory block at the very end
        free_shared_memory(shm_ext)
        release_shared_memory(shm_ext)
        return D

    def _fill_D_elements(self, D: npt.NDArray[np.float_], i: int) -> None:  # noqa: N802
        """
        fill in the D matrix elements using symmetries
        """
        D[i][i] = self._integral(i, i)
        for j in range(i + 1, self.mesh.mesh_eigenvalues.shape[0]):
            D[j][i] = self._integral(j, i)

    def _integral(self, i: int, j: int) -> float:
        """
        calculates the D integral between two mesh basis functions
        """
        return integrate_region_mesh(
            self.mesh.region,
            self.mesh.vertices,
            self.mesh.faces,
            self.mesh.basis_functions[i],
            self.mesh.basis_functions[j],
        )

    @staticmethod
    def _clean_evals_and_evecs(
        eigendecomposition: tuple,
    ) -> tuple[npt.NDArray[np.float_], npt.NDArray[np.float_]]:
        """
        need eigenvalues and eigenvectors to be in a certain format
        """
        # access values
        eigenvalues, eigenvectors = eigendecomposition

        # sort eigenvalues and eigenvectors in descending order of eigenvalues
        idx = eigenvalues.argsort()[::-1]
        eigenvalues = eigenvalues[idx]
        eigenvectors = eigenvectors[:, idx].T
        return eigenvalues, eigenvectors



import numpy as np
from numpy import typing as npt
from pydantic import validator
from pydantic.dataclasses import dataclass

from sleplet import logger
from sleplet.meshes.mesh_harmonic_coefficients import MeshHarmonicCoefficients
from sleplet.utils.harmonic_methods import mesh_forward
from sleplet.utils.validation import Validation


@dataclass(config=Validation, kw_only=True)
class MeshBasisFunctions(MeshHarmonicCoefficients):
    rank: int = 0

    def __post_init_post_parse__(self) -> None:
        self._validate_rank()
        super().__post_init_post_parse__()

    def _create_coefficients(self) -> npt.NDArray[np.complex_ | np.float_]:
        """
        compute field on the vertices of the mesh
        """
        logger.info(
            f"Mesh eigenvalue {self.rank}: "
            f"{self.mesh.mesh_eigenvalues[self.rank]:e}"
        )
        basis_function = self.mesh.basis_functions[self.rank]
        return mesh_forward(self.mesh, basis_function)

    def _create_name(self) -> str:
        return (
            (
                f"{self.mesh.name}_rank{self.rank}_"
                f"lam{self.mesh.mesh_eigenvalues[self.rank]:e}"
            )
            .replace(".", "-")
            .replace("+", "")
        )

    def _setup_args(self) -> None:
        if isinstance(self.extra_args, list):
            num_args = 1
            if len(self.extra_args) != num_args:
                raise ValueError(
                    f"The number of extra arguments should be 1 or {num_args}"
                )
            self.rank = self.extra_args[0]

    def _validate_rank(self) -> None:
        """
        checks the requested rank is valid
        """
        if isinstance(self.extra_args, list):
            limit = self.mesh.mesh_eigenvalues.shape[0]
            if self.extra_args[0] > limit:
                raise ValueError(f"rank should be less than or equal to {limit}")

    @validator("rank")
    def check_rank(cls, v):
        if not isinstance(v, int):
            raise TypeError("rank should be an integer")
        if v < 0:
            raise ValueError("rank cannot be negative")
        return v



from pathlib import Path

import numpy as np
import seaborn as sns
from matplotlib import pyplot as plt
from scipy.interpolate import pchip

from sleplet.utils.plot_methods import save_plot
from sleplet.utils.wavelet_methods import create_kappas

fig_path = Path(__file__).resolve().parents[2] / "src" / "sleplet" / "figures"
sns.set(context="paper")

B = 3
J_MIN = 2
L = 128
STEP = 0.01


def main() -> None:
    """
    plots the tiling of the harmonic line
    """
    xlim = L
    x = np.arange(xlim)
    xi = np.arange(0, xlim - 1 + STEP, STEP)
    kappas = create_kappas(xlim, B, J_MIN)
    yi = pchip(x, kappas[0])
    plt.semilogx(xi, yi(xi), label=r"$\Phi_{\ell0}$")
    for j, k in enumerate(kappas[1:]):
        yi = pchip(x, k)
        plt.semilogx(xi, yi(xi), label=rf"$\Psi^{{{j+J_MIN}}}_{{\ell0}}$")
    plt.xlim(1, xlim)
    ticks = 2 ** np.arange(np.log2(xlim) + 1, dtype=int)
    plt.xticks(ticks, ticks)
    plt.xlabel(r"$\ell$")
    plt.legend(loc=6)
    save_plot(fig_path, f"axisymmetric_tiling_L{L}")


if __name__ == "__main__":
    main()



import numpy as np
import pyssht as ssht
from numpy import typing as npt

from sleplet import logger
from sleplet.meshes.classes.mesh_slepian import MeshSlepian
from sleplet.meshes.classes.mesh_slepian_decomposition import MeshSlepianDecomposition
from sleplet.slepian.slepian_decomposition import SlepianDecomposition
from sleplet.slepian.slepian_functions import SlepianFunctions
from sleplet.slepian.slepian_region.slepian_arbitrary import SlepianArbitrary
from sleplet.slepian.slepian_region.slepian_limit_lat_lon import SlepianLimitLatLon
from sleplet.slepian.slepian_region.slepian_polar_cap import SlepianPolarCap
from sleplet.utils.harmonic_methods import boost_coefficient_resolution, mesh_inverse
from sleplet.utils.region import Region
from sleplet.utils.vars import SAMPLING_SCHEME


def choose_slepian_method(L: int, region: Region) -> SlepianFunctions:
    """
    initialise Slepian object depending on input
    """
    match region.region_type:
        case "polar":
            logger.info("polar cap region detected")
            return SlepianPolarCap(L, region.theta_max, gap=region.gap)

        case "lim_lat_lon":
            logger.info("limited latitude longitude region detected")
            return SlepianLimitLatLon(
                L,
                theta_min=region.theta_min,
                theta_max=region.theta_max,
                phi_min=region.phi_min,
                phi_max=region.phi_max,
            )

        case "arbitrary":
            logger.info("mask specified in file detected")
            return SlepianArbitrary(L, region.mask_name)

        case _:
            raise ValueError(f"{region.region_type} is an invalid region type")


def slepian_inverse(
    f_p: npt.NDArray[np.complex_ | np.float_], L: int, slepian: SlepianFunctions
) -> npt.NDArray[np.complex_]:
    """
    computes the Slepian inverse transform up to the Shannon number
    """
    f_p_reshape = f_p[: slepian.N, np.newaxis, np.newaxis]
    s_p = compute_s_p_omega(L, slepian)
    return (f_p_reshape * s_p).sum(axis=0)


def slepian_forward(
    L: int,
    slepian: SlepianFunctions,
    *,
    f: npt.NDArray[np.complex_] | None = None,
    flm: npt.NDArray[np.complex_ | np.float_] | None = None,
    mask: npt.NDArray[np.float_] | None = None,
    n_coeffs: int | None = None,
) -> npt.NDArray[np.complex_]:
    """
    computes the Slepian forward transform for all coefficients
    """
    sd = SlepianDecomposition(L, slepian, f=f, flm=flm, mask=mask)
    n_coeffs = slepian.N if n_coeffs is None else n_coeffs
    return sd.decompose_all(n_coeffs)


def compute_s_p_omega(L: int, slepian: SlepianFunctions) -> npt.NDArray[np.complex_]:
    """
    method to calculate Sp(omega) for a given region
    """
    n_theta, n_phi = ssht.sample_shape(L, Method=SAMPLING_SCHEME)
    sp = np.zeros((slepian.N, n_theta, n_phi), dtype=np.complex_)
    for p in range(slepian.N):
        if p % L == 0:
            logger.info(f"compute Sp(omega) p={p+1}/{slepian.N}")
        sp[p] = ssht.inverse(slepian.eigenvectors[p], L, Method=SAMPLING_SCHEME)
    return sp


def compute_s_p_omega_prime(
    L: int, alpha: float, beta: float, slepian: SlepianFunctions
) -> npt.NDArray[np.complex_]:
    """
    method to pick out the desired angle from Sp(omega)
    """
    sp_omega = compute_s_p_omega(L, slepian)
    p = ssht.theta_to_index(beta, L, Method=SAMPLING_SCHEME)
    q = ssht.phi_to_index(alpha, L, Method=SAMPLING_SCHEME)
    sp_omega_prime = sp_omega[:, p, q]
    # pad with zeros so it has the expected shape
    boost = L**2 - slepian.N
    return boost_coefficient_resolution(sp_omega_prime, boost)


def slepian_mesh_forward(
    mesh_slepian: MeshSlepian,
    *,
    u: npt.NDArray[np.complex_ | np.float_] | None = None,
    u_i: npt.NDArray[np.complex_ | np.float_] | None = None,
    mask: bool = False,
    n_coeffs: int | None = None,
) -> npt.NDArray[np.float_]:
    """
    computes the Slepian forward transform for all coefficients
    """
    sd = MeshSlepianDecomposition(
        mesh_slepian,
        u=u,
        u_i=u_i,
        mask=mask,
    )
    n_coeffs = mesh_slepian.N if n_coeffs is None else n_coeffs
    return sd.decompose_all(n_coeffs)


def slepian_mesh_inverse(
    mesh_slepian: MeshSlepian,
    f_p: npt.NDArray[np.complex_ | np.float_],
) -> npt.NDArray[np.complex_ | np.float_]:
    """
    computes the Slepian inverse transform on the mesh up to the Shannon number
    """
    f_p_reshape = f_p[: mesh_slepian.N, np.newaxis]
    s_p = compute_mesh_s_p_pixel(mesh_slepian)
    return (f_p_reshape * s_p).sum(axis=0)


def compute_mesh_s_p_pixel(mesh_slepian: MeshSlepian) -> npt.NDArray[np.float_]:
    """
    method to calculate Sp(omega) for a given region
    """
    sp = np.zeros((mesh_slepian.N, mesh_slepian.mesh.vertices.shape[0]))
    for p in range(mesh_slepian.N):
        sp[p] = mesh_inverse(mesh_slepian.mesh, mesh_slepian.slepian_functions[p])
    return sp



from concurrent.futures import ThreadPoolExecutor
from dataclasses import KW_ONLY
from pathlib import Path

import numpy as np
import pyssht as ssht
from numpy import linalg as LA  # noqa: N812
from numpy import typing as npt
from pydantic.dataclasses import dataclass

from sleplet import logger
from sleplet.data.setup_pooch import find_on_pooch_then_local
from sleplet.slepian.slepian_functions import SlepianFunctions
from sleplet.utils.array_methods import fill_upper_triangle_of_hermitian_matrix
from sleplet.utils.harmonic_methods import create_spherical_harmonic, invert_flm_boosted
from sleplet.utils.integration_methods import (
    calc_integration_weight,
    integrate_region_sphere,
)
from sleplet.utils.mask_methods import create_mask_region
from sleplet.utils.parallel_methods import (
    attach_to_shared_memory_block,
    create_shared_memory_array,
    free_shared_memory,
    release_shared_memory,
    split_arr_into_chunks,
)
from sleplet.utils.region import Region
from sleplet.utils.slepian_arbitrary_methods import clean_evals_and_evecs
from sleplet.utils.validation import Validation
from sleplet.utils.vars import NCPU, SAMPLES

_data_path = Path(__file__).resolve().parents[2] / "data"


@dataclass(config=Validation)
class SlepianArbitrary(SlepianFunctions):
    mask_name: str
    _: KW_ONLY

    def __post_init_post_parse__(self) -> None:
        self.resolution = SAMPLES * self.L
        super().__post_init_post_parse__()

    def _create_fn_name(self) -> str:
        return f"slepian_{self.mask_name}"

    def _create_region(self) -> Region:
        return Region(mask_name=self.mask_name)

    def _create_mask(self) -> npt.NDArray[np.float_]:
        return create_mask_region(self.resolution, self.region)

    def _calculate_area(self) -> float:
        self.weight = calc_integration_weight(self.resolution)
        return (self.mask * self.weight).sum()

    def _create_matrix_location(self) -> str:
        return f"slepian_eigensolutions_D_{self.mask_name}_L{self.L}_N{self.N}"

    def _solve_eigenproblem(
        self,
    ) -> tuple[npt.NDArray[np.float_], npt.NDArray[np.complex_]]:
        eval_loc = f"{self.matrix_location}_eigenvalues.npy"
        evec_loc = f"{self.matrix_location}_eigenvectors.npy"

        try:
            return np.load(find_on_pooch_then_local(eval_loc)), np.load(
                find_on_pooch_then_local(evec_loc)
            )
        except TypeError:
            return self._solve_D_matrix(eval_loc, evec_loc)

    def _solve_D_matrix(  # noqa: N802
        self, eval_loc, evec_loc
    ) -> tuple[npt.NDArray[np.float_], npt.NDArray[np.complex_]]:
        D = self._create_D_matrix()

        # fill in remaining triangle section
        fill_upper_triangle_of_hermitian_matrix(D)

        # solve eigenproblem
        eigenvalues, eigenvectors = clean_evals_and_evecs(LA.eigh(D))
        np.save(_data_path / eval_loc, eigenvalues)
        np.save(_data_path / evec_loc, eigenvectors[: self.N])
        return eigenvalues, eigenvectors

    def _create_D_matrix(self) -> npt.NDArray[np.complex_]:  # noqa: N802
        """
        computes the D matrix in parallel
        """
        # create dictionary for the integrals
        self._fields: dict[int, npt.NDArray[np.complex_ | np.float_]] = {}

        # initialise real and imaginary matrices
        D_r = np.zeros((self.L**2, self.L**2))
        D_i = np.zeros((self.L**2, self.L**2))

        D_r_ext, shm_r_ext = create_shared_memory_array(D_r)
        D_i_ext, shm_i_ext = create_shared_memory_array(D_i)

        def func(chunk: list[int]) -> None:
            """
            calculate D matrix components for each chunk
            """
            D_r_int, shm_r_int = attach_to_shared_memory_block(D_r, shm_r_ext)
            D_i_int, shm_i_int = attach_to_shared_memory_block(D_i, shm_i_ext)

            for i in chunk:
                logger.info(f"start ell: {i}")
                self._matrix_helper(D_r_int, D_i_int, i)
                logger.info(f"finish ell: {i}")

            free_shared_memory(shm_r_int, shm_i_int)

        # split up L range to maximise effiency
        chunks = split_arr_into_chunks(self.L**2, NCPU)

        # initialise pool and apply function
        with ThreadPoolExecutor(max_workers=NCPU) as e:
            e.map(func, chunks)

        # retrieve from parallel function
        D = D_r_ext + 1j * D_i_ext

        # Free and release the shared memory block at the very end
        free_shared_memory(shm_r_ext, shm_i_ext)
        release_shared_memory(shm_r_ext, shm_i_ext)
        return D

    def _matrix_helper(
        self, D_r: npt.NDArray[np.float_], D_i: npt.NDArray[np.float_], i: int
    ) -> None:
        """
        used in both serial and parallel calculations

        the hack with splitting into real and imaginary parts
        is not required for the serial case but here for ease
        """
        # fill in diagonal components
        integral = self._integral(i, i)
        D_r[i][i] = integral.real
        D_i[i][i] = integral.imag
        _, m_i = ssht.ind2elm(i)

        for j in range(i + 1, D_r.shape[0]):
            ell_j, m_j = ssht.ind2elm(j)
            # if possible to use previous calculations
            if m_i == 0 and m_j != 0 and ell_j < self.L:
                # if positive m then use conjugate relation
                if m_j > 0:
                    integral = self._integral(j, i)
                    D_r[j][i] = integral.real
                    D_i[j][i] = integral.imag
                    k = ssht.elm2ind(ell_j, -m_j)
                    D_r[k][i] = (-1) ** m_j * D_r[j][i]
                    D_i[k][i] = (-1) ** (m_j + 1) * D_i[j][i]
            else:
                integral = self._integral(j, i)
                D_r[j][i] = integral.real
                D_i[j][i] = integral.imag

    def _integral(self, i: int, j: int) -> complex:
        """
        calculates the D integral between two spherical harmonics
        """
        if i not in self._fields:
            self._fields[i] = invert_flm_boosted(
                create_spherical_harmonic(self.L, i), self.L, self.resolution
            )
        if j not in self._fields:
            self._fields[j] = invert_flm_boosted(
                create_spherical_harmonic(self.L, j), self.L, self.resolution
            )
        return integrate_region_sphere(
            self.mask, self.weight, self._fields[i], self._fields[j].conj()
        )



from abc import abstractmethod
from dataclasses import KW_ONLY

import numpy as np
from numpy import typing as npt
from pydantic import validator
from pydantic.dataclasses import dataclass

from sleplet.utils.convolution_methods import sifting_convolution
from sleplet.utils.mask_methods import ensure_masked_flm_bandlimited
from sleplet.utils.region import Region
from sleplet.utils.string_methods import filename_args
from sleplet.utils.validation import Validation

COEFFICIENTS_TO_NOT_MASK: set[str] = {"slepian", "south", "america"}


@dataclass(config=Validation)
class Coefficients:
    L: int
    _: KW_ONLY
    extra_args: list[int] | None = None
    noise: float | None = None
    region: Region | None = None
    smoothing: int | None = None

    def __post_init_post_parse__(self) -> None:
        self._setup_args()
        self.name = self._create_name()
        self.spin = self._set_spin()
        self.reality = self._set_reality()
        self.coefficients = self._create_coefficients()
        self._add_details_to_name()
        self.unnoised_coefficients, self.snr = self._add_noise_to_signal()

    def translate(
        self, alpha: float, beta: float, *, shannon: int | None = None
    ) -> npt.NDArray[np.complex_ | np.float_]:
        g_coefficients = self._translation_helper(alpha, beta)
        return (
            g_coefficients
            if "dirac_delta" in self.name
            else self.convolve(self.coefficients, g_coefficients, shannon=shannon)
        )

    def convolve(
        self,
        f_coefficient: npt.NDArray[np.complex_ | np.float_],
        g_coefficient: npt.NDArray[np.complex_ | np.float_],
        *,
        shannon: int | None = None,
    ) -> npt.NDArray[np.complex_ | np.float_]:
        # translation/convolution are not real for general function
        self.reality = False
        return sifting_convolution(f_coefficient, g_coefficient, shannon=shannon)

    def _add_details_to_name(self) -> None:
        """
        adds region to the name if present if not a Slepian function
        adds noise/smoothing if appropriate and bandlimit
        """
        if (
            isinstance(self.region, Region)
            and not set(self.name.split("_")) & COEFFICIENTS_TO_NOT_MASK
        ):
            self.name += f"_{self.region.name_ending}"
        if self.noise is not None:
            self.name += f"{filename_args(self.noise, 'noise')}"
        if self.smoothing is not None:
            self.name += f"{filename_args(self.smoothing, 'smoothed')}"
        self.name += f"_L{self.L}"

    @validator("coefficients", check_fields=False)
    def check_coefficients(cls, v, values):
        if (
            values["region"]
            and not set(values["name"].split("_")) & COEFFICIENTS_TO_NOT_MASK
        ):
            v = ensure_masked_flm_bandlimited(
                v,
                values["L"],
                values["region"],
                reality=values["reality"],
                spin=values["spin"],
            )
        return v

    @abstractmethod
    def rotate(
        self, alpha: float, beta: float, *, gamma: float = 0
    ) -> npt.NDArray[np.complex_]:
        """
        rotates given flm on the sphere by alpha/beta/gamma
        """
        raise NotImplementedError

    @abstractmethod
    def _translation_helper(
        self, alpha: float, beta: float
    ) -> npt.NDArray[np.complex_]:
        """
        compute the basis function at omega' for translation
        """
        raise NotImplementedError

    @abstractmethod
    def _add_noise_to_signal(
        self,
    ) -> tuple[npt.NDArray[np.complex_ | np.float_] | None, float | None]:
        """
        adds Gaussian white noise to the signal
        """
        raise NotImplementedError

    @abstractmethod
    def _create_coefficients(self) -> npt.NDArray[np.complex_ | np.float_]:
        """
        creates the flm on the north pole
        """
        raise NotImplementedError

    @abstractmethod
    def _create_name(self) -> str:
        """
        creates the name of the function
        """
        raise NotImplementedError

    @abstractmethod
    def _set_reality(self) -> bool:
        """
        sets the reality flag to speed up computations
        """
        raise NotImplementedError

    @abstractmethod
    def _set_spin(self) -> int:
        """
        sets the spin value in computations
        """
        raise NotImplementedError

    @abstractmethod
    def _setup_args(self) -> None:
        """
        initialises function specific args
        either default value or user input
        """
        raise NotImplementedError
